@@ -1,7 +1,6 @@
 /* GNU m4 -- A simple macro processor
 
-   Copyright (C) 1989, 1990, 1991, 1992, 1993, 1994, 2004, 2005, 2006,
-   2007, 2008 Free Software Foundation, Inc.
+   Copyright (C) 1989-1994, 2004-2013 Free Software Foundation, Inc.
 
    This file is part of GNU M4.
 
@@ -23,6 +22,8 @@
 
 #include "m4.h"
 
+#include "memchr2.h"
+
 /* Unread input can be either files, that should be read (eg. included
    files), strings, which should be rescanned (eg. macro expansion text),
    or quoted macro definitions (as returned by the builtin "defn").
@@ -64,35 +65,36 @@
 
 enum input_type
 {
-  INPUT_STRING,		/* String resulting from macro expansion.  */
-  INPUT_FILE,		/* File from command line or include.  */
-  INPUT_MACRO		/* Builtin resulting from defn.  */
+  INPUT_STRING,         /* String resulting from macro expansion.  */
+  INPUT_FILE,           /* File from command line or include.  */
+  INPUT_MACRO           /* Builtin resulting from defn.  */
 };
 
 typedef enum input_type input_type;
 
 struct input_block
 {
-  struct input_block *prev;	/* previous input_block on the input stack */
-  input_type type;		/* see enum values */
-  const char *file;		/* file where this input is from */
-  int line;			/* line where this input is from */
+  struct input_block *prev;     /* previous input_block on the input stack */
+  input_type type;              /* see enum values */
+  const char *file;             /* file where this input is from */
+  int line;                     /* line where this input is from */
   union
     {
       struct
-	{
-	  char *string;		/* remaining string value */
-	}
-	u_s;	/* INPUT_STRING */
+        {
+          char *string;         /* remaining string value */
+          char *end;            /* terminating NUL of string */
+        }
+        u_s;    /* INPUT_STRING */
       struct
-	{
-	  FILE *fp;		     /* input file handle */
-	  bool_bitfield end : 1;     /* true if peek has seen EOF */
-	  bool_bitfield close : 1;   /* true if we should close file on pop */
-	  bool_bitfield advance : 1; /* track previous start_of_input_line */
-	}
-	u_f;	/* INPUT_FILE */
-      builtin_func *func;	/* pointer to macro's function */
+        {
+          FILE *fp;                  /* input file handle */
+          bool_bitfield end : 1;     /* true if peek has seen EOF */
+          bool_bitfield close : 1;   /* true if we should close file on pop */
+          bool_bitfield advance : 1; /* track previous start_of_input_line */
+        }
+        u_f;    /* INPUT_FILE */
+      builtin_func *func;       /* pointer to macro's function */
     }
   u;
 };
@@ -136,8 +138,8 @@ static bool start_of_input_line;
 /* Flag for next_char () to recognize change in input block.  */
 static bool input_change;
 
-#define CHAR_EOF	256	/* character return on EOF */
-#define CHAR_MACRO	257	/* character return for MACRO token */
+#define CHAR_EOF        256     /* character return on EOF */
+#define CHAR_MACRO      257     /* character return for MACRO token */
 
 /* Quote chars.  */
 STRING rquote;
@@ -151,7 +153,6 @@ STRING ecomm;
 
 # define DEFAULT_WORD_REGEXP "[_a-zA-Z][_a-zA-Z0-9]*"
 
-static char *word_start;
 static struct re_pattern_buffer word_regexp;
 static int default_word_regexp;
 static struct re_registers regs;
@@ -188,7 +189,7 @@ push_file (FILE *fp, const char *title, bool close_when_done)
     DEBUG_MESSAGE1 ("input read from %s", title);
 
   i = (input_block *) obstack_alloc (current_input,
-				     sizeof (struct input_block));
+                                     sizeof (struct input_block));
   i->type = INPUT_FILE;
   i->file = (char *) obstack_copy0 (&file_names, title, strlen (title));
   i->line = 1;
@@ -222,7 +223,7 @@ push_macro (builtin_func *func)
     }
 
   i = (input_block *) obstack_alloc (current_input,
-				     sizeof (struct input_block));
+                                     sizeof (struct input_block));
   i->type = INPUT_MACRO;
   i->file = current_file;
   i->line = current_line;
@@ -235,7 +236,7 @@ push_macro (builtin_func *func)
 
 /*------------------------------------------------------------------.
 | First half of push_string ().  The pointer next points to the new |
-| input_block.							    |
+| input_block.                                                      |
 `------------------------------------------------------------------*/
 
 struct obstack *
@@ -244,12 +245,12 @@ push_string_init (void)
   if (next != NULL)
     {
       M4ERROR ((warning_status, 0,
-		"INTERNAL ERROR: recursive push_string!"));
+                "INTERNAL ERROR: recursive push_string!"));
       abort ();
     }
 
   next = (input_block *) obstack_alloc (current_input,
-					sizeof (struct input_block));
+                                        sizeof (struct input_block));
   next->type = INPUT_STRING;
   next->file = current_file;
   next->line = current_line;
@@ -257,14 +258,15 @@ push_string_init (void)
   return current_input;
 }
 
-/*------------------------------------------------------------------------.
-| Last half of push_string ().  If next is now NULL, a call to push_file  |
-| () has invalidated the previous call to push_string_init (), so we just |
-| give up.  If the new object is void, we do not push it.  The function	  |
-| push_string_finish () returns a pointer to the finished object.  This	  |
-| pointer is only for temporary use, since reading the next token might	  |
-| release the memory used for the object.				  |
-`------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| Last half of push_string ().  If next is now NULL, a call to       |
+| push_file () has invalidated the previous call to push_string_init |
+| (), so we just give up.  If the new object is void, we do not push |
+| it.  The function push_string_finish () returns a pointer to the   |
+| finished object.  This pointer is only for temporary use, since    |
+| reading the next token might release the memory used for the       |
+| object.                                                            |
+`-------------------------------------------------------------------*/
 
 const char *
 push_string_finish (void)
@@ -276,11 +278,13 @@ push_string_finish (void)
 
   if (obstack_object_size (current_input) > 0)
     {
+      size_t len = obstack_object_size (current_input);
       obstack_1grow (current_input, '\0');
       next->u.u_s.string = (char *) obstack_finish (current_input);
+      next->u.u_s.end = next->u.u_s.string + len;
       next->prev = isp;
       isp = next;
-      ret = isp->u.u_s.string;	/* for immediate use only */
+      ret = isp->u.u_s.string; /* for immediate use only */
       input_change = true;
     }
   else
@@ -301,23 +305,26 @@ push_string_finish (void)
 void
 push_wrapup (const char *s)
 {
+  size_t len = strlen (s);
   input_block *i;
   i = (input_block *) obstack_alloc (wrapup_stack,
-				     sizeof (struct input_block));
+                                     sizeof (struct input_block));
   i->prev = wsp;
   i->type = INPUT_STRING;
   i->file = current_file;
   i->line = current_line;
-  i->u.u_s.string = (char *) obstack_copy0 (wrapup_stack, s, strlen (s));
+  i->u.u_s.string = (char *) obstack_copy0 (wrapup_stack, s, len);
+  i->u.u_s.end = i->u.u_s.string + len;
   wsp = i;
 }
 
 
-/*-------------------------------------------------------------------------.
-| The function pop_input () pops one level of input sources.  If the	   |
-| popped input_block is a file, current_file and current_line are reset to |
-| the saved values before the memory for the input_block are released.	   |
-`-------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| The function pop_input () pops one level of input sources.  If the |
+| popped input_block is a file, current_file and current_line are    |
+| reset to the saved values before the memory for the input_block is |
+| released.                                                          |
+`-------------------------------------------------------------------*/
 
 static void
 pop_input (void)
@@ -332,47 +339,48 @@ pop_input (void)
 
     case INPUT_FILE:
       if (debug_level & DEBUG_TRACE_INPUT)
-	{
-	  if (tmp)
-	    DEBUG_MESSAGE2 ("input reverted to %s, line %d",
-			    tmp->file, tmp->line);
-	  else
-	    DEBUG_MESSAGE ("input exhausted");
-	}
+        {
+          if (tmp)
+            DEBUG_MESSAGE2 ("input reverted to %s, line %d",
+                            tmp->file, tmp->line);
+          else
+            DEBUG_MESSAGE ("input exhausted");
+        }
 
       if (ferror (isp->u.u_f.fp))
-	{
-	  M4ERROR ((warning_status, 0, "read error"));
-	  if (isp->u.u_f.close)
-	    fclose (isp->u.u_f.fp);
-	  retcode = EXIT_FAILURE;
-	}
+        {
+          M4ERROR ((warning_status, 0, "read error"));
+          if (isp->u.u_f.close)
+            fclose (isp->u.u_f.fp);
+          retcode = EXIT_FAILURE;
+        }
       else if (isp->u.u_f.close && fclose (isp->u.u_f.fp) == EOF)
-	{
-	  M4ERROR ((warning_status, errno, "error reading file"));
-	  retcode = EXIT_FAILURE;
-	}
+        {
+          M4ERROR ((warning_status, errno, "error reading file"));
+          retcode = EXIT_FAILURE;
+        }
       start_of_input_line = isp->u.u_f.advance;
       output_current_line = -1;
       break;
 
     default:
       M4ERROR ((warning_status, 0,
-		"INTERNAL ERROR: input stack botch in pop_input ()"));
+                "INTERNAL ERROR: input stack botch in pop_input ()"));
       abort ();
     }
   obstack_free (current_input, isp);
-  next = NULL;			/* might be set in push_string_init () */
+  next = NULL; /* might be set in push_string_init () */
 
   isp = tmp;
   input_change = true;
 }
 
-/*------------------------------------------------------------------------.
-| To switch input over to the wrapup stack, main () calls pop_wrapup ().  |
-| Since wrapup text can install new wrapup text, pop_wrapup () returns	  |
-| false when there is no wrapup text on the stack, and true otherwise.	  |
-`------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| To switch input over to the wrapup stack, main calls pop_wrapup    |
+| ().  Since wrapup text can install new wrapup text, pop_wrapup ()  |
+| returns false when there is no wrapup text on the stack, and true  |
+| otherwise.                                                         |
+`-------------------------------------------------------------------*/
 
 bool
 pop_wrapup (void)
@@ -384,11 +392,14 @@ pop_wrapup (void)
   if (wsp == NULL)
     {
       /* End of the program.  Free all memory even though we are about
-	 to exit, since it makes leak detection easier.  */
+         to exit, since it makes leak detection easier.  */
       obstack_free (&token_stack, NULL);
       obstack_free (&file_names, NULL);
       obstack_free (wrapup_stack, NULL);
       free (wrapup_stack);
+#ifdef ENABLE_CHANGEWORD
+      regfree (&word_regexp);
+#endif /* ENABLE_CHANGEWORD */
       return false;
     }
 
@@ -414,7 +425,7 @@ init_macro_token (token_data *td)
   if (isp->type != INPUT_MACRO)
     {
       M4ERROR ((warning_status, 0,
-		"INTERNAL ERROR: bad call to init_macro_token ()"));
+                "INTERNAL ERROR: bad call to init_macro_token ()"));
       abort ();
     }
 
@@ -423,12 +434,12 @@ init_macro_token (token_data *td)
 }
 
 
-/*------------------------------------------------------------------------.
-| Low level input is done a character at a time.  The function peek_input |
-| () is used to look at the next character in the input stream.  At any	  |
-| given time, it reads from the input_block on the top of the current	  |
-| input stack.								  |
-`------------------------------------------------------------------------*/
+/*-----------------------------------------------------------------.
+| Low level input is done a character at a time.  The function     |
+| peek_input () is used to look at the next character in the input |
+| stream.  At any given time, it reads from the input_block on the |
+| top of the current input stack.                                  |
+`-----------------------------------------------------------------*/
 
 static int
 peek_input (void)
@@ -439,52 +450,52 @@ peek_input (void)
   while (1)
     {
       if (block == NULL)
-	return CHAR_EOF;
+        return CHAR_EOF;
 
       switch (block->type)
-	{
-	case INPUT_STRING:
-	  ch = to_uchar (block->u.u_s.string[0]);
-	  if (ch != '\0')
-	    return ch;
-	  break;
+        {
+        case INPUT_STRING:
+          ch = to_uchar (block->u.u_s.string[0]);
+          if (ch != '\0')
+            return ch;
+          break;
 
-	case INPUT_FILE:
-	  ch = getc (block->u.u_f.fp);
-	  if (ch != EOF)
-	    {
-	      ungetc (ch, block->u.u_f.fp);
-	      return ch;
-	    }
-	  block->u.u_f.end = true;
-	  break;
+        case INPUT_FILE:
+          ch = getc (block->u.u_f.fp);
+          if (ch != EOF)
+            {
+              ungetc (ch, block->u.u_f.fp);
+              return ch;
+            }
+          block->u.u_f.end = true;
+          break;
 
-	case INPUT_MACRO:
-	  return CHAR_MACRO;
+        case INPUT_MACRO:
+          return CHAR_MACRO;
 
-	default:
-	  M4ERROR ((warning_status, 0,
-		    "INTERNAL ERROR: input stack botch in peek_input ()"));
-	  abort ();
-	}
+        default:
+          M4ERROR ((warning_status, 0,
+                    "INTERNAL ERROR: input stack botch in peek_input ()"));
+          abort ();
+        }
       block = block->prev;
     }
 }
 
-/*-------------------------------------------------------------------------.
-| The function next_char () is used to read and advance the input to the   |
-| next character.  It also manages line numbers for error messages, so	   |
-| they do not get wrong, due to lookahead.  The token consisting of a	   |
-| newline alone is taken as belonging to the line it ends, and the current |
-| line number is not incremented until the next character is read.	   |
-| 99.9% of all calls will read from a string, so factor that out into a    |
-| macro for speed.                                                         |
-`-------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| The function next_char () is used to read and advance the input to |
+| the next character.  It also manages line numbers for error        |
+| messages, so they do not get wrong, due to lookahead.  The token   |
+| consisting of a newline alone is taken as belonging to the line it |
+| ends, and the current line number is not incremented until the     |
+| next character is read.  99.9% of all calls will read from a       |
+| string, so factor that out into a macro for speed.                 |
+`-------------------------------------------------------------------*/
 
 #define next_char() \
-  (isp && isp->type == INPUT_STRING && isp->u.u_s.string[0]	\
-   && !input_change						\
-   ? to_uchar (*isp->u.u_s.string++)				\
+  (isp && isp->type == INPUT_STRING && isp->u.u_s.string[0]     \
+   && !input_change                                             \
+   ? to_uchar (*isp->u.u_s.string++)                            \
    : next_char_1 ())
 
 static int
@@ -495,66 +506,65 @@ next_char_1 (void)
   while (1)
     {
       if (isp == NULL)
-	{
-	  current_file = "";
-	  current_line = 0;
-	  return CHAR_EOF;
-	}
+        {
+          current_file = "";
+          current_line = 0;
+          return CHAR_EOF;
+        }
 
       if (input_change)
-	{
-	  current_file = isp->file;
-	  current_line = isp->line;
-	  input_change = false;
-	}
+        {
+          current_file = isp->file;
+          current_line = isp->line;
+          input_change = false;
+        }
 
       switch (isp->type)
-	{
-	case INPUT_STRING:
-	  ch = to_uchar (*isp->u.u_s.string++);
-	  if (ch != '\0')
-	    return ch;
-	  break;
+        {
+        case INPUT_STRING:
+          ch = to_uchar (*isp->u.u_s.string++);
+          if (ch != '\0')
+            return ch;
+          break;
 
-	case INPUT_FILE:
-	  if (start_of_input_line)
-	    {
-	      start_of_input_line = false;
-	      current_line = ++isp->line;
-	    }
+        case INPUT_FILE:
+          if (start_of_input_line)
+            {
+              start_of_input_line = false;
+              current_line = ++isp->line;
+            }
 
-	  /* If stdin is a terminal, calling getc after peek_input
-	     already called it would make the user have to hit ^D
-	     twice to quit.  */
-	  ch = isp->u.u_f.end ? EOF : getc (isp->u.u_f.fp);
-	  if (ch != EOF)
-	    {
-	      if (ch == '\n')
-		start_of_input_line = true;
-	      return ch;
-	    }
-	  break;
+          /* If stdin is a terminal, calling getc after peek_input
+             already called it would make the user have to hit ^D
+             twice to quit.  */
+          ch = isp->u.u_f.end ? EOF : getc (isp->u.u_f.fp);
+          if (ch != EOF)
+            {
+              if (ch == '\n')
+                start_of_input_line = true;
+              return ch;
+            }
+          break;
 
-	case INPUT_MACRO:
-	  pop_input ();		/* INPUT_MACRO input sources has only one
-				   token */
-	  return CHAR_MACRO;
+        case INPUT_MACRO:
+          pop_input (); /* INPUT_MACRO input sources has only one token */
+          return CHAR_MACRO;
 
-	default:
-	  M4ERROR ((warning_status, 0,
-		    "INTERNAL ERROR: input stack botch in next_char ()"));
-	  abort ();
-	}
+        default:
+          M4ERROR ((warning_status, 0,
+                    "INTERNAL ERROR: input stack botch in next_char ()"));
+          abort ();
+        }
 
       /* End of input source --- pop one level.  */
       pop_input ();
     }
 }
 
-/*------------------------------------------------------------------------.
-| skip_line () simply discards all immediately following characters, upto |
-| the first newline.  It is only used from m4_dnl ().			  |
-`------------------------------------------------------------------------*/
+/*-------------------------------------------------------------------.
+| skip_line () simply discards all immediately following characters, |
+| upto the first newline.  It is only used from m4_dnl ().           |
+`-------------------------------------------------------------------*/
 
 void
 skip_line (void)
@@ -569,7 +579,7 @@ skip_line (void)
     /* current_file changed to "" if we see CHAR_EOF, use the
        previous value we stored earlier.  */
     M4ERROR_AT_LINE ((warning_status, 0, file, line,
-		      "Warning: end of file treated as newline"));
+                      "Warning: end of file treated as newline"));
   /* On the rare occasion that dnl crosses include file boundaries
      (either the input file did not end in a newline, or changeword
      was used), calling next_char can update current_file and
@@ -591,34 +601,34 @@ skip_line (void)
 static bool
 match_input (const char *s, bool consume)
 {
-  int n;			/* number of characters matched */
-  int ch;			/* input character */
+  int n;                        /* number of characters matched */
+  int ch;                       /* input character */
   const char *t;
   bool result = false;
 
   ch = peek_input ();
   if (ch != to_uchar (*s))
-    return false;			/* fail */
+    return false;                       /* fail */
 
   if (s[1] == '\0')
     {
       if (consume)
-	(void) next_char ();
-      return true;			/* short match */
+        next_char ();
+      return true;                      /* short match */
     }
 
-  (void) next_char ();
-  for (n = 1, t = s++; (ch = peek_input ()) == to_uchar (*s++); )
+  next_char ();
+  for (n = 1, t = s++; peek_input () == to_uchar (*s++); )
     {
-      (void) next_char ();
+      next_char ();
       n++;
-      if (*s == '\0')		/* long match */
-	{
-	  if (consume)
-	    return true;
-	  result = true;
-	  break;
-	}
+      if (*s == '\0')           /* long match */
+        {
+          if (consume)
+            return true;
+          result = true;
+          break;
+        }
     }
 
   /* Failed or shouldn't consume, push back input.  */
@@ -694,7 +704,7 @@ input_init (void)
 
 
 /*------------------------------------------------------------------.
-| Functions for setting quotes and comment delimiters.  Used by	    |
+| Functions for setting quotes and comment delimiters.  Used by     |
 | m4_changecom () and m4_changequote ().  Pass NULL if the argument |
 | was not present, to distinguish from an explicit empty string.    |
 `------------------------------------------------------------------*/
@@ -755,12 +765,10 @@ set_comment (const char *bc, const char *ec)
 void
 set_word_regexp (const char *regexp)
 {
-  int i;
-  char test[2];
   const char *msg;
   struct re_pattern_buffer new_word_regexp;
 
-  if (!*regexp || !strcmp (regexp, DEFAULT_WORD_REGEXP))
+  if (!*regexp || STREQ (regexp, DEFAULT_WORD_REGEXP))
     {
       default_word_regexp = true;
       return;
@@ -774,35 +782,23 @@ set_word_regexp (const char *regexp)
   if (msg != NULL)
     {
       M4ERROR ((warning_status, 0,
-		"bad regular expression `%s': %s", regexp, msg));
+                "bad regular expression `%s': %s", regexp, msg));
       return;
     }
 
-  /* If compilation worked, retry using the word_regexp struct.
-     Can't rely on struct assigns working, so redo the compilation.  */
-  regfree (&word_regexp);
+  /* If compilation worked, retry using the word_regexp struct.  We
+     can't rely on struct assigns working, so redo the compilation.
+     The fastmap can be reused between compilations, and will be freed
+     by the final regfree.  */
+  if (!word_regexp.fastmap)
+    word_regexp.fastmap = xcharalloc (UCHAR_MAX + 1);
   msg = re_compile_pattern (regexp, strlen (regexp), &word_regexp);
+  assert (!msg);
   re_set_registers (&word_regexp, &regs, regs.num_regs, regs.start, regs.end);
-
-  if (msg != NULL)
-    {
-      M4ERROR ((EXIT_FAILURE, 0,
-		"INTERNAL ERROR: expression recompilation `%s': %s",
-		regexp, msg));
-    }
+  if (re_compile_fastmap (&word_regexp))
+    assert (false);
 
   default_word_regexp = false;
-
-  if (word_start == NULL)
-    word_start = (char *) xmalloc (256);
-
-  word_start[0] = '\0';
-  test[1] = '\0';
-  for (i = 1; i < 256; i++)
-    {
-      test[0] = i;
-      word_start[i] = re_search (&word_regexp, test, 1, 0, 0, NULL) >= 0;
-    }
 }
 
 #endif /* ENABLE_CHANGEWORD */
@@ -856,7 +852,7 @@ next_token (token_data *td, int *line)
       next_char ();
 #ifdef DEBUG_INPUT
       xfprintf (stderr, "next_token -> MACDEF (%s)\n",
-		find_builtin_by_addr (TOKEN_DATA_FUNC (td))->name);
+                find_builtin_by_addr (TOKEN_DATA_FUNC (td))->name);
 #endif
       return TOKEN_MACDEF;
     }
@@ -868,15 +864,15 @@ next_token (token_data *td, int *line)
     {
       obstack_grow (&token_stack, bcomm.string, bcomm.length);
       while ((ch = next_char ()) != CHAR_EOF
-	     && !MATCH (ch, ecomm.string, true))
-	obstack_1grow (&token_stack, ch);
+             && !MATCH (ch, ecomm.string, true))
+        obstack_1grow (&token_stack, ch);
       if (ch != CHAR_EOF)
-	obstack_grow (&token_stack, ecomm.string, ecomm.length);
+        obstack_grow (&token_stack, ecomm.string, ecomm.length);
       else
-	/* current_file changed to "" if we see CHAR_EOF, use the
-	   previous value we stored earlier.  */
-	M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, *line,
-			  "ERROR: end of file in comment"));
+        /* current_file changed to "" if we see CHAR_EOF, use the
+           previous value we stored earlier.  */
+        M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, *line,
+                          "ERROR: end of file in comment"));
 
       type = TOKEN_STRING;
     }
@@ -884,46 +880,46 @@ next_token (token_data *td, int *line)
     {
       obstack_1grow (&token_stack, ch);
       while ((ch = peek_input ()) != CHAR_EOF && (isalnum (ch) || ch == '_'))
-	{
-	  obstack_1grow (&token_stack, ch);
-	  (void) next_char ();
-	}
+        {
+          obstack_1grow (&token_stack, ch);
+          next_char ();
+        }
       type = TOKEN_WORD;
     }
 
 #ifdef ENABLE_CHANGEWORD
 
-  else if (!default_word_regexp && word_start[ch])
+  else if (!default_word_regexp && word_regexp.fastmap[ch])
     {
       obstack_1grow (&token_stack, ch);
       while (1)
-	{
-	  ch = peek_input ();
-	  if (ch == CHAR_EOF)
-	    break;
-	  obstack_1grow (&token_stack, ch);
-	  startpos = re_search (&word_regexp,
-				(char *) obstack_base (&token_stack),
-				obstack_object_size (&token_stack), 0, 0,
-				&regs);
-	  if (startpos != 0 ||
-	      regs.end [0] != obstack_object_size (&token_stack))
-	    {
-	      *(((char *) obstack_base (&token_stack)
-		 + obstack_object_size (&token_stack)) - 1) = '\0';
-	      break;
-	    }
-	  next_char ();
-	}
+        {
+          ch = peek_input ();
+          if (ch == CHAR_EOF)
+            break;
+          obstack_1grow (&token_stack, ch);
+          startpos = re_search (&word_regexp,
+                                (char *) obstack_base (&token_stack),
+                                obstack_object_size (&token_stack), 0, 0,
+                                &regs);
+          if (startpos ||
+              regs.end [0] != (regoff_t) obstack_object_size (&token_stack))
+            {
+              *(((char *) obstack_base (&token_stack)
+                 + obstack_object_size (&token_stack)) - 1) = '\0';
+              break;
+            }
+          next_char ();
+        }
 
       obstack_1grow (&token_stack, '\0');
       orig_text = (char *) obstack_finish (&token_stack);
 
       if (regs.start[1] != -1)
-	obstack_grow (&token_stack,orig_text + regs.start[1],
-		      regs.end[1] - regs.start[1]);
+        obstack_grow (&token_stack,orig_text + regs.start[1],
+                      regs.end[1] - regs.start[1]);
       else
-	obstack_grow (&token_stack, orig_text,regs.end[0]);
+        obstack_grow (&token_stack, orig_text,regs.end[0]);
 
       type = TOKEN_WORD;
     }
@@ -933,48 +929,85 @@ next_token (token_data *td, int *line)
   else if (!MATCH (ch, lquote.string, true))
     {
       switch (ch)
-	{
-	case '(':
-	  type = TOKEN_OPEN;
-	  break;
-	case ',':
-	  type = TOKEN_COMMA;
-	  break;
-	case ')':
-	  type = TOKEN_CLOSE;
-	  break;
-	default:
-	  type = TOKEN_SIMPLE;
-	  break;
-	}
+        {
+        case '(':
+          type = TOKEN_OPEN;
+          break;
+        case ',':
+          type = TOKEN_COMMA;
+          break;
+        case ')':
+          type = TOKEN_CLOSE;
+          break;
+        default:
+          type = TOKEN_SIMPLE;
+          break;
+        }
       obstack_1grow (&token_stack, ch);
     }
   else
     {
+      bool fast = lquote.length == 1 && rquote.length == 1;
       quote_level = 1;
       while (1)
-	{
-	  ch = next_char ();
-	  if (ch == CHAR_EOF)
-	    /* current_file changed to "" if we see CHAR_EOF, use
-	       the previous value we stored earlier.  */
-	    M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, *line,
-			      "ERROR: end of file in string"));
+        {
+          /* Try scanning a buffer first.  */
+          const char *buffer = (isp && isp->type == INPUT_STRING
+                                ? isp->u.u_s.string : NULL);
+          if (buffer && *buffer)
+            {
+              size_t len = isp->u.u_s.end - buffer;
+              const char *p = buffer;
+              do
+                {
+                  p = (char *) memchr2 (p, *lquote.string, *rquote.string,
+                                        buffer + len - p);
+                }
+              while (p && fast && (*p++ == *rquote.string
+                                   ? --quote_level : ++quote_level));
+              if (p)
+                {
+                  if (fast)
+                    {
+                      assert (!quote_level);
+                      obstack_grow (&token_stack, buffer, p - buffer - 1);
+                      isp->u.u_s.string += p - buffer;
+                      break;
+                    }
+                  obstack_grow (&token_stack, buffer, p - buffer);
+                  ch = to_uchar (*p);
+                  isp->u.u_s.string += p - buffer + 1;
+                }
+              else
+                {
+                  obstack_grow (&token_stack, buffer, len);
+                  isp->u.u_s.string += len;
+                  continue;
+                }
+            }
+          /* Fall back to a byte.  */
+          else
+            ch = next_char ();
+          if (ch == CHAR_EOF)
+            /* current_file changed to "" if we see CHAR_EOF, use
+               the previous value we stored earlier.  */
+            M4ERROR_AT_LINE ((EXIT_FAILURE, 0, file, *line,
+                              "ERROR: end of file in string"));
 
-	  if (MATCH (ch, rquote.string, true))
-	    {
-	      if (--quote_level == 0)
-		break;
-	      obstack_grow (&token_stack, rquote.string, rquote.length);
-	    }
-	  else if (MATCH (ch, lquote.string, true))
-	    {
-	      quote_level++;
-	      obstack_grow (&token_stack, lquote.string, lquote.length);
-	    }
-	  else
-	    obstack_1grow (&token_stack, ch);
-	}
+          if (MATCH (ch, rquote.string, true))
+            {
+              if (--quote_level == 0)
+                break;
+              obstack_grow (&token_stack, rquote.string, rquote.length);
+            }
+          else if (MATCH (ch, lquote.string, true))
+            {
+              quote_level++;
+              obstack_grow (&token_stack, lquote.string, lquote.length);
+            }
+          else
+            obstack_1grow (&token_stack, ch);
+        }
       type = TOKEN_STRING;
     }
 
@@ -989,7 +1022,7 @@ next_token (token_data *td, int *line)
 #endif
 #ifdef DEBUG_INPUT
   xfprintf (stderr, "next_token -> %s (%s)\n",
-	    token_type_string (type), TOKEN_DATA_TEXT (td));
+            token_type_string (type), TOKEN_DATA_TEXT (td));
 #endif
   return type;
 }
@@ -1018,9 +1051,9 @@ peek_token (void)
     }
   else if ((default_word_regexp && (isalpha (ch) || ch == '_'))
 #ifdef ENABLE_CHANGEWORD
-      || (! default_word_regexp && word_start[ch])
+           || (! default_word_regexp && word_regexp.fastmap[ch])
 #endif /* ENABLE_CHANGEWORD */
-      )
+           )
     {
       result = TOKEN_WORD;
     }
@@ -1032,16 +1065,16 @@ peek_token (void)
     switch (ch)
       {
       case '(':
-	result = TOKEN_OPEN;
-	break;
+        result = TOKEN_OPEN;
+        break;
       case ',':
-	result = TOKEN_COMMA;
-	break;
+        result = TOKEN_COMMA;
+        break;
       case ')':
-	result = TOKEN_CLOSE;
-	break;
+        result = TOKEN_CLOSE;
+        break;
       default:
-	result = TOKEN_SIMPLE;
+        result = TOKEN_SIMPLE;
       }
 
 #ifdef DEBUG_INPUT
@@ -1057,7 +1090,7 @@ static const char *
 token_type_string (token_type t)
 {
  switch (t)
-    {				/* TOKSW */
+    { /* TOKSW */
     case TOKEN_EOF:
       return "EOF";
     case TOKEN_STRING:
@@ -1084,7 +1117,7 @@ print_token (const char *s, token_type t, token_data *td)
 {
   xfprintf (stderr, "%s: ", s);
   switch (t)
-    {				/* TOKSW */
+    { /* TOKSW */
     case TOKEN_OPEN:
     case TOKEN_COMMA:
     case TOKEN_CLOSE:
